#include "main_code.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include "include/Engine.h"

#include "main_code.h"

TMyApplication* Application;

Matrix3f quatToMatrix(Vector4f q) {
	Matrix3f result;

	double sqw = q(3)*q(3);
	double sqx = q(0)*q(0);
	double sqy = q(1)*q(1);
	double sqz = q(2)*q(2);

	// invs (inverse square length) is only required if quaternion is not already normalised
	double invs = 1 / (sqx + sqy + sqz + sqw);

	result(0,0) = (sqx - sqy - sqz + sqw)*invs; // since sqw + sqx + sqy + sqz =1/invs*invs
	result(1,1) = (-sqx + sqy - sqz + sqw)*invs;
	result(2,2) = (-sqx - sqy + sqz + sqw)*invs;

	double tmp1 = q(0)*q(1);
	double tmp2 = q(2)*q(3);
	result(1, 0) = 2.0 * (tmp1 + tmp2)*invs;
	result(0, 1) = 2.0 * (tmp1 - tmp2)*invs;

	tmp1 = q(0)*q(2);
	tmp2 = q(1)*q(3);
	result(2,0) = 2.0 * (tmp1 - tmp2)*invs;
	result(0,2) = 2.0 * (tmp1 + tmp2)*invs;
	tmp1 = q(1)*q(2);
	tmp2 = q(0)*q(3);
	result(2, 1) = 2.0 * (tmp1 + tmp2)*invs;
	result(1, 2) = 2.0 * (tmp1 - tmp2)*invs;

	return result;
}


void TMyApplication::InnerInit() {
    Application = this;
    
#ifdef TARGET_WIN32
#ifdef NDEBUG
	ST::PathToResources = "../../../assets/";
	//ST::PathToResources = "resources/";
#else
	ST::PathToResources = "../../../assets/";
#endif
#endif
    
#ifdef TARGET_IOS
    ST::PathToResources = "assets/";
#endif

    if (Console != NULL) {
        *Console<<"APP INIT\n";
    }
    srand (static_cast<size_t>(time(NULL)));

	ResourceManager->ShaderManager.AddShader("DefaultShader", "shader1vertex.txt", "shader1fragment.txt");
	ResourceManager->ShaderManager.AddShader("FrameShader", "frameshader_vertex.txt", "frameshader_fragment.txt");
	ResourceManager->ShaderManager.AddShader("ColorShader", "color_vertex.txt", "color_fragment.txt");
	ResourceManager->ShaderManager.AddShader("SSAA_4X", "SSAA_4X.vertex", "SSAA_4X.frag");
	ResourceManager->ShaderManager.AddShader("ParallaxShader", "parallax_vertex.txt", "parallax_fragment.txt");
	ResourceManager->ShaderManager.AddShader("Transparency", "Transparency.vertex", "Transparency.fragment");
	ResourceManager->ShaderManager.AddShader("TransparencyPlusColor", "transparencyPlusColor.vertex", "transparencyPlusColor.fragment");
	Renderer->PushShader("DefaultShader");

	ResourceManager->TexList.AddTexture("console_bkg.bmp");
	ResourceManager->TexList.AddTexture("owl-green.jpg");
	ResourceManager->TexList.AddTexture("owl-green-height.png");
	ResourceManager->TexList.AddTexture("owl-green-normal.png");

	ResourceManager->TexList.AddTexture("pimgpsh.jpg");
	ResourceManager->TexList.AddTexture("fabric-texture.jpg");
	ResourceManager->TexList.AddTexture("fabric-normal-map.png");
	ResourceManager->TexList.AddTexture("fabric-height-map.png");
	ResourceManager->TexList.AddTexture("fabric-transparency-mask.png");

	ResourceManager->TexList.AddTexture("5f.jpg");
	ResourceManager->TexList.AddTexture("6f.jpg");
	ResourceManager->TexList.AddTexture("7f.jpg");

	{
		rect.second.Data.Vec3CoordArr[CONST_STRING_POSITION_ATTRIB].push_back(Vector3f(-4096, 0, -4096));
		rect.second.Data.Vec3CoordArr[CONST_STRING_POSITION_ATTRIB].push_back(Vector3f(-4096, 0, 4096));
		rect.second.Data.Vec3CoordArr[CONST_STRING_POSITION_ATTRIB].push_back(Vector3f(4096, 0, 4096));
		rect.second.Data.Vec3CoordArr[CONST_STRING_POSITION_ATTRIB].push_back(Vector3f(-4096, 0, -4096));
		rect.second.Data.Vec3CoordArr[CONST_STRING_POSITION_ATTRIB].push_back(Vector3f(4096, 0, 4096));
		rect.second.Data.Vec3CoordArr[CONST_STRING_POSITION_ATTRIB].push_back(Vector3f(4096, 0, -4096));

		rect.second.Data.Vec2CoordArr[CONST_STRING_TEXCOORD_ATTRIB].push_back(Vector2f(0, 0));
		rect.second.Data.Vec2CoordArr[CONST_STRING_TEXCOORD_ATTRIB].push_back(Vector2f(0, 1));
		rect.second.Data.Vec2CoordArr[CONST_STRING_TEXCOORD_ATTRIB].push_back(Vector2f(1, 1));
		rect.second.Data.Vec2CoordArr[CONST_STRING_TEXCOORD_ATTRIB].push_back(Vector2f(0, 0));
		rect.second.Data.Vec2CoordArr[CONST_STRING_TEXCOORD_ATTRIB].push_back(Vector2f(1, 1));
		rect.second.Data.Vec2CoordArr[CONST_STRING_TEXCOORD_ATTRIB].push_back(Vector2f(1, 0));

		rect.first.SamplerMap[CONST_STRING_TEXTURE_UNIFORM] = "pimgpsh.jpg";
		rect.second.RefreshBuffer();
	}

	{
		size_t const mapWidth = 2048;
		size_t const mapHeight = 2048;
		size_t const channelsCount = 4;
		char *heightMap = new char[mapWidth * mapHeight * channelsCount];
		char *material = new char[mapWidth * mapHeight * channelsCount];
		char *normalMap = new char[mapWidth * mapHeight * channelsCount];

		for(size_t z = 0; z < mapHeight; z++) {
			for(size_t x = 0; x < mapWidth; x++) {
				size_t index = channelsCount * (z * mapWidth + x);

				heightMap[index + 0] = 0;
				heightMap[index + 1] = 0;
				heightMap[index + 2] = 0;
				heightMap[index + 3] = 255;

				material[index + 0] = 255;
				material[index + 1] = 255;
				material[index + 2] = 255;
				material[index + 3] = 0;

				normalMap[index + 0] = 128;
				normalMap[index + 1] = 128;
				normalMap[index + 2] = 255;
				normalMap[index + 3] = 255;
			}
		}

		double *heightMapF = new double[mapWidth * mapHeight];
		const size_t threadsCount = 3;
		const double threadWidth = 4.0;
		const auto g = [this, mapWidth, mapHeight, channelsCount, threadWidth, threadsCount, heightMapF, material, normalMap](Vector2d start, Vector2d end, Vector3f &color) {
			start = start + Vector2d(mapWidth / 2, mapHeight / 2);
			end = end + Vector2d(mapWidth / 2, mapHeight / 2);

			const auto length = (end - start).norm();
			const Vector2d e0 = (end - start).normalized();
			const Vector2d e1(-e0.y(), e0.x());

			std::vector<Vector2d> corners;
			corners.push_back(start + threadWidth * e1);
			corners.push_back(start - threadWidth * e1);
			corners.push_back(end + threadWidth * e1);
			corners.push_back(end - threadWidth * e1);

			auto minXf = corners[0].x();
			auto maxXf = corners[0].x();
			auto minYf = corners[0].y();
			auto maxYf = corners[0].y();

			for(auto corner: corners) {
				if(corner.x() < minXf) {minXf = corner.x();}
				if(corner.x() > maxXf) {maxXf = corner.x();}
				if(corner.y() < minYf) {minYf = corner.y();}
				if(corner.y() > maxYf) {maxYf = corner.y();}
			}

			const auto minX = floor(minXf);
			const auto maxX = ceil(maxXf);
			const auto minY = floor(minYf);
			const auto maxY = floor(maxYf);

			const double r = threadWidth / 2;
			Matrix2d basisMatrix;
			basisMatrix << e0.x(), e0.y(), e1.x(), e1.y();
			for(auto x = minX; x <= maxX; x++) {
				for(auto y = minY; y <= maxY; y++) {
					if(x < 0 || x >= mapWidth || y < 0 || y >= mapHeight) {continue;}
					Vector2d self = basisMatrix * Vector2d(x - start.x(), y - start.y());

					if(self.x() >= 0 && self.x() <= length && self.y() >= -threadWidth && self.y() <= threadWidth) {
						const double phase = 6 * pi * self.x() / length;
						std::vector<Vector2d> centers;
						
						auto maxH = -1.0f;
						Vector3d normal;
						for(size_t i = 0; i < threadsCount; i++) {
							Vector2d center = 
								r * sin(phase + 2 * pi * i / threadsCount) * Vector2d(1, 0) +
								r * cos(phase + 2 * pi * i / threadsCount) * Vector2d(0, 1);

							auto temp = r * r - (self.y() - center.x()) * (self.y() - center.x());
							if(temp < 0) {continue;}

							float h = sqrt(temp) + center.y();
							if(h > maxH) {
								maxH = h;
								Vector3d center3d =
									self.x() * Vector3d(e0.x(), e0.y(), 0) +
									center.x() * Vector3d(e1.x(), e1.y(), 0) +
									center.y() * Vector3d(0, 0, 1);
								Vector3d intersection3d =
									self.x() * Vector3d(e0.x(), e0.y(), 0) +
									self.y() * Vector3d(e1.x(), e1.y(), 0) +
									maxH * Vector3d(0, 0, 1);
								normal = (intersection3d - center3d).normalized();
							}
						}

						size_t index = y * mapWidth + x;
						if(heightMapF[index] > maxH / (2 * r)) {continue;}

						heightMapF[index] = maxH / (2 * r);
						material[4 * index] = 255 * color(0);
						material[4 * index + 1] = 255 * color(1);
						material[4 * index + 2] = 255 * color(2);
						material[4 * index + 3] = 255;

						normalMap[4 * index] = 255 * (normal.x() + 1) / 2;
						normalMap[4 * index + 1] = 255 * (normal.y() + 1) / 2;
						normalMap[4 * index + 2] = 255 * normal.z();
						normalMap[4 * index + 3] = 255;
					}
				}
			}
		};

		namespace pt = boost::property_tree;
		pt::ptree root;
		pt::read_json(ST::PathToResources + "lines.json", root);
		size_t counter = 0;
		for (auto line : root.get_child("lines")) {
			std::vector<double> start;
			std::vector<double> end;
			std::vector<double> color;
			for (auto value : line.second.get_child("start")) {
				start.push_back(value.second.get_value<double>());
			}
			for (auto value : line.second.get_child("end")) {
				end.push_back(value.second.get_value<double>());
			}
			for (auto value : line.second.get_child("color")) {
				color.push_back(value.second.get_value<double>());
			}
			g(Vector2d(start[0] / 4, start[1] / 4), Vector2d(end[0] / 4, end[1] / 4), Vector3f(color[0], color[1], color[2]));
			counter++;
			//if(counter > 3000) {
			//	break;
			//}
		}

		for(size_t i = 0; i < mapWidth; i++) {
			for(size_t j = 0; j < mapHeight; j++) {
				heightMap[4 * (i * mapWidth + j)] = 255 - 255 * heightMapF[i * mapWidth + j];
				heightMap[4 * (i * mapWidth + j) + 1] = 255 - 255 * heightMapF[i * mapWidth + j];
				heightMap[4 * (i * mapWidth + j) + 2] = 255 - 255 * heightMapF[i * mapWidth + j];
				heightMap[4 * (i * mapWidth + j) + 3] = 255;
			}
		}
		delete[] heightMapF;

		SE::TTextureData heightMapTexture;
		heightMapTexture.Width = mapWidth;
		heightMapTexture.Height = mapHeight;
		heightMapTexture.Format = "bmp32";
		heightMapTexture.DataSize = mapWidth * mapHeight * channelsCount;
		heightMapTexture.Data = boost::shared_array<char>(heightMap);
		ResourceManager->TexList.AddTexture("HeightMapTexture", heightMapTexture);


		SE::TTextureData materialTexture;
		materialTexture.Width = mapWidth;
		materialTexture.Height = mapHeight;
		materialTexture.Format = "bmp32";
		materialTexture.DataSize = mapWidth * mapHeight * channelsCount;
		materialTexture.Data = boost::shared_array<char>(material);
		ResourceManager->TexList.AddTexture("MaterialTexture", materialTexture);

		SE::TTextureData normalMapTexture;
		normalMapTexture.Width = mapWidth;
		normalMapTexture.Height = mapHeight;
		normalMapTexture.Format = "bmp32";
		normalMapTexture.DataSize = mapWidth * mapHeight * channelsCount;
		normalMapTexture.Data = boost::shared_array<char>(normalMap);
		ResourceManager->TexList.AddTexture("NormalMapTexture", normalMapTexture);
	}

	{
		pair.second.Data.Vec3CoordArr[CONST_STRING_POSITION_ATTRIB].push_back(Vector3f(-2048, 5, -2048));
		pair.second.Data.Vec3CoordArr[CONST_STRING_POSITION_ATTRIB].push_back(Vector3f(-2048, 5, 2048));
		pair.second.Data.Vec3CoordArr[CONST_STRING_POSITION_ATTRIB].push_back(Vector3f(2048, 5, 2048));
		pair.second.Data.Vec3CoordArr[CONST_STRING_POSITION_ATTRIB].push_back(Vector3f(-2048, 5, -2048));
		pair.second.Data.Vec3CoordArr[CONST_STRING_POSITION_ATTRIB].push_back(Vector3f(2048, 5, 2048));
		pair.second.Data.Vec3CoordArr[CONST_STRING_POSITION_ATTRIB].push_back(Vector3f(2048, 5, -2048));

		pair.second.Data.Vec2CoordArr[CONST_STRING_TEXCOORD_ATTRIB].push_back(Vector2f(0, 0));
		pair.second.Data.Vec2CoordArr[CONST_STRING_TEXCOORD_ATTRIB].push_back(Vector2f(0, 1));
		pair.second.Data.Vec2CoordArr[CONST_STRING_TEXCOORD_ATTRIB].push_back(Vector2f(1, 1));
		pair.second.Data.Vec2CoordArr[CONST_STRING_TEXCOORD_ATTRIB].push_back(Vector2f(0, 0));
		pair.second.Data.Vec2CoordArr[CONST_STRING_TEXCOORD_ATTRIB].push_back(Vector2f(1, 1));
		pair.second.Data.Vec2CoordArr[CONST_STRING_TEXCOORD_ATTRIB].push_back(Vector2f(1, 0));

		pair.second.RefreshBuffer();
		pair.first.ShaderName = "ParallaxShader";

		pair.first.SamplerMap[CONST_STRING_TEXTURE_UNIFORM] = "MaterialTexture";
		pair.first.SamplerMap["HeightMap"] = "HeightMapTexture";
		pair.first.SamplerMap["NormalMap"] = "NormalMapTexture";
		pair.first.SamplerMap["TransparencyMask"] = "MaterialTexture";
	}

	Inited = true;
}

void TMyApplication::InnerDeinit()
{
    Inited = false;
    Loaded = false;
    if (Console != NULL)
    {
        *Console<<"APP DEINIT\n";
    }
}

void TMyApplication::InnerOnTapDown(Vector2f p) {}

void TMyApplication::InnerOnTapUp(Vector2f p) {}

void TMyApplication::InnerOnTapUpAfterMove(Vector2f p) {}

void TMyApplication::InnerOnMove(Vector2f p, Vector2f shift)
{
	phi += shift(1)*0.02f;

	if (phi < pi/12)
	{
		phi = pi / 12;
	}

	if (phi > pi / 2)
	{
		phi = pi / 2;
	}

	alpha -= shift(0)*0.02f;
}
	
void TMyApplication::OnFling(Vector2f v) 
{
}

void TMyApplication::OnMouseWheel(short int delta)
{
	distance += delta;
}
	
void TMyApplication::InnerDraw()
{
	Renderer->SwitchToScreen();
	Renderer->SetPerspectiveProjection(pi / 6, 10.f, 10000.f);
	Renderer->SetFullScreenViewport();

	Renderer->PushMatrix();

	Renderer->TranslateMatrix(Vector3f(0, 0, -distance));
	Vector4f quat1 = Vector4f(sin(phi / 2), 0, 0, cos(phi / 2));
	Vector4f quat2 = Vector4f(0, sin(alpha / 2), 0, cos(alpha / 2));
	Renderer->RotateMatrix(quat1);
	Renderer->RotateMatrix(quat2);

	glDisable(GL_DEPTH_TEST);
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
	glClear(GL_COLOR_BUFFER_BIT);
	CheckGlError("");

	auto mat1 = quatToMatrix(quat1);
	auto mat2 = quatToMatrix(quat2);
	Vector3f lightPos = {0.f, 1.f, 1.f};
	Vector3f eye = mat2 * (mat1 * Vector3f(0.0f, 0.f, -distance));

	{
		TRenderParamsSetter params(rect.first);
		Renderer->DrawTriangleList(rect.second);
	}

	{
		TRenderParamsSetter params(pair.first);
		RenderUniform3fv("eye", eye.data());
		RenderUniform3fv("lightPos", lightPos.data());
		Matrix3f normMatrix = Renderer->GetModelviewMatrix().inverse().transpose().block<3, 3>(0, 0);

		RenderUniformMatrix3fv("NormalMatrix", false, normMatrix.data());
		RenderUniformMatrix4fv("ModelViewMatrix", false, Renderer->GetModelviewMatrix().data());
		RenderUniformMatrix3fv("ModelViewMatrix3x3", false, Renderer->GetModelviewMatrix().block<3, 3>(0, 0).data());
		Renderer->DrawTriangleList(pair.second);
	}

	Renderer->PopMatrix();
	CheckGlError("");
}

void TMyApplication::InnerUpdate(size_t dt) {}

bool TMyApplication::IsLoaded() {
    return Loaded;
}

bool TMyApplication::IsInited() {
    return Inited;
}
